/*
** File: adb3_platform.c  
** Project: ADB3 API module library
** Purpose: Implements VxWorks-specific functions for ADB3 API module library
**
** (C) Copyright Alpha Data 2011
**
** Mechanism for reporting errors from driver to API when calling ioctl() is as follows:
**
**   - A successful ioctl() call always returns OK.
**   - An error reported by the driver framework results in ioctl() returning ERROR, and the driver setting errno appropriately. See below for
**     errno usage.
**   - An error reported by the platform-independent code results in ioctl() returning a positive number, which is then cast to ADB3_STATUS.
**
** Error numbers (errno) are used as follows:
**
**   EACCES      (O)    - Insufficient rights to open device; shouldn't happen in VxWorks
**   ENXIO       (O)    - Device not found (i.e. /adb3/<index> doesn't exist
**   ENDEV       (O)    - As ENXIO
**   ENOMEM      (O, I) - Couldn't allocate memory, either from heap or from a pool in order to perform operation
**   EBADF       (I, C) - File descriptor was not valid
**   EFAULT      (I)    - The IOCTL structure passed was NULL, or the in/out size encoded in the IOCTL number was not what the driver expected.
**   ENOSYS      (I)    - The IOCTL number passed was not one recognized by the driver
**   EBUSY       (I)    - There is already a non-blocking operation in progress
**   EINPROGRESS (I)    - The non-blocking operation is still ongoing
**   EPIPE       (I)    - No non-blocking operation to finish
**
** where O, I, and C mean applicable to open(), ioctl(), and close() respectively.
*/

#include <ioLib.h>
#include <selectLib.h>
#include <assert.h>
#include <errno.h>

#include "adb3_platform.h"
#include <ioctl_dfvxw.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static const char g_deviceName[] = "adb3";

ADB3_STATUS
adb3OpenDevice(
  unsigned int index,
  bool_t bPassive, /* ignored in VxWorks */
  ADB3_HANDLE* phadb3)
{
  char filename[32];
  int hDevice;

#if (_WRS_VXWORKS_MAJOR >= 6)
  snprintf(filename, ARRAY_LENGTH(filename), "/%s/%u", g_deviceName, index);
#else
  /* Need VxWorks 6.x for snprintf() */
  sprintf(filename, "/%s/%u", g_deviceName, index);
#endif
  hDevice = open(filename, O_RDWR, 0);
  if (ERROR == hDevice) {
    switch (errno) {
    case EACCES:
      return ADB3_ACCESS_DENIED;

    case ENXIO:
    case ENODEV:
      return ADB3_DEVICE_NOT_FOUND;

    case ENOMEM:
      return ADB3_NO_MEMORY;

    default:
      fprintf(stderr, "adb3OpenDevice: errno=%d (%s)\n", errno, strerror(errno));
      return ADB3_UNEXPECTED_ERROR;
    }
  }
  *phadb3 = hDevice;
  return ADB3_SUCCESS;
}

ADB3_STATUS
adb3CloseDevice(
  ADB3_HANDLE hadb3)
{
  if (ERROR == close(hadb3)) {
    switch (errno) {
      case EBADF:
        return ADB3_INVALID_HANDLE;

      default:
        fprintf(stderr, "adb3CloseDevice: errno=%d (%s)\n", errno, strerror(errno));
        return ADB3_UNEXPECTED_ERROR;
    }
  }    

  return ADB3_SUCCESS;
}

ADB3_STATUS
adb3Ioctl(
  ADB3_HANDLE hadb3,
  unsigned int code,
  void* pIoctl,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  STATUS err;

  if (sizeIn >= DFVXW_MAX_IOCTL_SIZE || sizeOut >= DFVXW_MAX_IOCTL_SIZE) {
    fprintf(stderr, "adb3Ioctl: sizeIn (%u) or sizeOut (%u) greater than 64\n", sizeIn, sizeOut);
    return ADB3_UNEXPECTED_ERROR;
  }
  err = ioctl(hadb3, DFVXW_MAKE_IOCTL(code, sizeIn, sizeOut), (int)pIoctl);
  if (OK == err) {
    return ADB3_SUCCESS;
  } else {
    if (ERROR == err) {
      switch (errno) {
      case EBADF:
        return ADB3_INVALID_HANDLE;

      case ENOMEM:
        return ADB3_NO_MEMORY;

      case ECANCELED:
        return ADB3_CANCELLED;

      case EFAULT:
      case ENOSYS:
      default:
        fprintf(stderr, "adb3Ioctl: errno=%d (%s)\n", errno, strerror(errno));
        return ADB3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADB3_STATUS)(err & 0xfffU);
    }
  }
}
