/*
** File: si5338.c  
** Project: ADB3 core driver
** Purpose: Functions for programming the SI5338 clock generator.
**
** (C) Copyright Alpha Data 2011
**
** SI5338 frequency formulae are as follows:
**
**   Fpfd = Fref / P1
**   Fvco = Fpfd * MSn
**   Foutx = Fvco / (MSx * Rx) = Fin * MSn / (P * MSx * Rx)
**
** where x is the multisynth index in the range 0..3.
**
** The driver does not change the PLL, input stage or output stage registers.
** The output stage registers are expected to be configured for divide by 1.
** In order to program a new frequency on one of the four clock outputs, the
** driver keeps the VCO frequency constant and changes only the multisynth
** registers for the clock output in question.
**
** The function si5338FullConfiguration is called when the driver detects,
** via the PCI revision ID, that the firmware doesn't do the full SI5338
** configuration at reset/power-up. In that case, the driver does the full
** SI5338 configuration.
**
** Regardless of whether or not si5338FullConfiguration is called, in
** si5338Interrogate, the driver finds the VCO frequency from the reference
** clock frequency and by reading the SI5338 registers. The calculated VCO
** frequency is then used as the basis for all multisynth-related
** calculations when changing the frequencies of the four clock outputs.
*/

#include <df.h>

#include "coreif.h"
#include "i2c_common.h"
#include "si5338_common.h"

/* From SI5338 data sheet */
static const uint32_t g_fVcoMin = 2200000000U;        /* Minimum permited VCO frequency */    
static const uint32_t g_fVcoMax = 2840000000U;        /* Maximum permited VCO frequency */

/*
** Register write masks for all SI5338 registers that apply ONLY when writing
** the ClockBuilder configuration to the SI5338. These masks should not be used
** in general when writing the SI5338.
**
** A 0 bit means that a bit's value must be preserved when its register is written.
*/
static const uint8_t g_registerMasks[351] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1D, 0x00,  /* 0x000 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x008 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x010 */
  0x00, 0x00, 0x00, 0x80, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x018 */
  0xFF, 0xFF, 0xFF, 0xFF, 0x1F, 0x1F, 0x1F, 0x1F,  /* 0x020 */
  0xFF, 0x7F, 0x3F, 0x00, 0x00, 0xFF, 0xFF, 0x3F,  /* 0x028 */
  0xFF, 0xFF, 0xFF, 0xFF, 0x7F, 0xFF, 0xFF, 0xFF,  /* 0x030 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x3F, 0x7F,  /* 0x038 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x040 */
  0xFF, 0x3F, 0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x048 */
  0xFF, 0xFF, 0xFF, 0xFF, 0x3F, 0x7F, 0xFF, 0xFF,  /* 0x050 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x3F,  /* 0x058 */
  0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x060 */
  0xFF, 0xFF, 0xBF, 0xFF, 0x7F, 0xFF, 0xFF, 0xFF,  /* 0x068 */
  0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x070 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x078 */
  0xFF, 0x0F, 0x0F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x080 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x088 */
  0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x090 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x0F, 0x0F,  /* 0x098 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0A0 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0A8 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x0F, 0xFF, 0xFF,  /* 0x0B0 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0B8 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0C0 */
  0xFF, 0xFF, 0xFF, 0x0F, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0C8 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0D0 */
  0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x0D8 */
  0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x0F, 0x00,  /* 0x0E0 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x0E8 */
  0x00, 0x0F, 0x02, 0x00, 0x00, 0x00, 0x0F, 0x00,  /* 0x0F0 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,  /* 0x0F8 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x100 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x108 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x110 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,  /* 0x118 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x120 */
  0xFF, 0xFF, 0xFF, 0x0F, 0x00, 0x00, 0x00, 0xFF,  /* 0x128 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x130 */
  0xFF, 0xFF, 0xFF, 0x0F, 0x00, 0x00, 0x00, 0xFF,  /* 0x138 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x140 */
  0xFF, 0xFF, 0xFF, 0x0F, 0x00, 0x00, 0x00, 0xFF,  /* 0x148 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x150 */
  0xFF, 0xFF, 0xFF, 0x0F, 0x00, 0x00, 0x00         /* 0x158 */
};

static Si5338Status
mapI2cStatus(
  I2cStatus status)
{
  switch (status) {
  case I2cStatusSuccess:
    return Si5338StatusSuccess;

  case I2cStatusTimeout:
    return Si5338StatusI2CTimeout;

  case I2cStatusHardwareError:
    return Si5338StatusI2CError;

  case I2cStatusInvalidBus:
  case I2cStatusInvalidDevice:
  case I2cStatusInvalidAddress:
  case I2cStatusGeneralFailure:
  default:
    return Si5338StatusGeneralFailure;
  }
}

Si5338ClockWordStatus
si5338ClockWord(
	uint32_t frequencyVCO,
	uint64_t frequencyReq,
	uint64_t frequencyMin,
	uint64_t frequencyMax,
  CoreClockWord* pClockWord)
{
  const uint32_t fOutMin = frequencyVCO / 568; /* Minimum possible output frequency = fVCO / 568 */
  const uint32_t fOutMax = frequencyVCO >> 3;  /* Maximum possible output frequency = fVCO / 8 */
  uint32_t msxaValue, msxbValue, msxcValue;    /* MSx value as (a + b/c) */
  uint64_t tmp;

	dfDebugPrint(3, ("si5338ClockWord: frequencyVCO=%lu reqFrequency=%llu\n",
    (unsigned long)frequencyVCO, (unsigned long long)frequencyReq));

  if (frequencyVCO < g_fVcoMin || frequencyVCO > g_fVcoMax) {
    /* VCO frequency should be in the range 2.2 to 2.84 GHz */
		return Si5338ClockWordStatusGeneralFailure;
  }

  if (frequencyReq < (fOutMin - 1000) || frequencyReq > (fOutMax + 1000)) {
		return Si5338ClockWordStatusOutOfRange;
  }

  /* Find MSx value, as (a + b/c) */
  msxaValue = frequencyVCO / (uint32_t)frequencyReq;
  if (msxaValue < 8U || msxaValue > 568U) {
    /* This means that the requested frequency is out of range */
		return Si5338ClockWordStatusOutOfRange;
  }
  msxbValue = frequencyVCO % (uint32_t)frequencyReq;
  msxcValue = (uint32_t)frequencyReq;

  /*
  ** Get MSx_P1, MSx_P2 and MSx_P3 from a, b and c values:
  **
  **   MSx_P1 = (128 * (a * c + b)) / c - 512
  **   MSx_P2 = (128 * (a * c + b)) % c
  **   MSx_P3 = c
  **
  ** where / is a division operator that rounds down to nearest integer
  **       % is a modulus operator
  */
  tmp = (uint64_t)128U * ((uint64_t)msxaValue * (uint64_t)msxcValue + (uint64_t)msxbValue);
  pClockWord->d0 /* MSx_P1 */ = (uint32_t)(dfDivide64By32Rem(tmp, msxcValue, &pClockWord->d1 /* MSx_P2 */)) - 512U;
  pClockWord->d2 /* MSx_P3 */ = msxcValue;                           

  /* Actual frequency will always be very close to requested frequency */
	pClockWord->frequency = frequencyReq;

	dfDebugPrint(3, ("si5338ClockWord: bestFrequency=%lu p1=0x%08lx p2=0x%08lx p3=0x%08lx\n",
    (unsigned long)frequencyReq, (unsigned long)pClockWord->d0, (unsigned long)pClockWord->d1, (unsigned long)pClockWord->d2));

	return Si5338ClockWordStatusSuccess;
}

static Si5338Status
si5338ReadByteSync(
  Si5338DeviceContext* pSi5338Ctx,
  uint8_t address,
  uint8_t* pVal8)
{
  I2cStatus status;

  status = i2cReadSync(pSi5338Ctx->pI2cCtx, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, address, pVal8);
  dfDebugPrint(7, ("si5338ReadByteSync: busIndex=0x%x slot=0x%02x address=0x%02x data=0x%02x\n", (unsigned int)pSi5338Ctx->busIndex, (unsigned int)pSi5338Ctx->deviceIndex, (unsigned int)address, (unsigned int)*pVal8));
  return mapI2cStatus(status);
}

static Si5338Status
si5338WriteByteSync(
  Si5338DeviceContext* pSi5338Ctx,
  uint8_t address,
  uint8_t val8)
{
  I2cStatus status;

  dfDebugPrint(7, ("si5338WriteByteSync: busIndex=0x%x slot=0x%02x address=0x%02x data=0x%02x\n", (unsigned int)pSi5338Ctx->busIndex, (unsigned int)pSi5338Ctx->deviceIndex, (unsigned int)address, (unsigned int)val8));
  status = i2cWriteSync(pSi5338Ctx->pI2cCtx, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, address, val8);
  return mapI2cStatus(status);
}

/* Must be called while holding page bit serializer */
static Si5338Status
si5338SetPageBit(
  Si5338DeviceContext* pSi5338Ctx,
  uint8_t page)
{
  Si5338Status status;
  uint8_t tmp8;

  status = si5338ReadByteSync(pSi5338Ctx, SI5338_REGIDX_PAGE, &tmp8);
  if (Si5338StatusSuccess != status) {
    return status;
  }
  tmp8 &= ~(uint8_t)0x1U;
  tmp8 |= page & (uint8_t)0x1U;
  return si5338WriteByteSync(pSi5338Ctx, SI5338_REGIDX_PAGE, tmp8);
}

Si5338Status
si5338ReadSync(
  Si5338DeviceContext* pSi5338Ctx,
  uint16_t regIndex,
  uint8_t* pVal8)
{
  SerializerRequestSync request;
  uint8_t page;
  Si5338Status status;

  if (regIndex >= DF_ARRAY_LENGTH(g_registerMasks)) {
    return Si5338StatusInvalidRegister;
  }
  page = (uint8_t)((regIndex >> 8) & 0x1U);
  serializerGetSync(&pSi5338Ctx->pageBitSerializer, &request);
  if (pSi5338Ctx->pageBit != page) {
    pSi5338Ctx->pageBit = page;
    status = si5338WriteByteSync(pSi5338Ctx, SI5338_REGIDX_PAGE, page);
    if (Si5338StatusSuccess != status) {
      /* Force page bit to be changed on next access */
      pSi5338Ctx->pageBit = 0xffU;
      serializerPut(&pSi5338Ctx->pageBitSerializer);
      return status;
    }
  }
  status = si5338ReadByteSync(pSi5338Ctx, (uint8_t)regIndex, pVal8);
  serializerPut(&pSi5338Ctx->pageBitSerializer);
  return status;
}

Si5338Status
si5338WriteMaskSync(
  Si5338DeviceContext* pSi5338Ctx,
  uint16_t regIndex,
  uint8_t mask8,
  uint8_t val8)
{
  SerializerRequestSync request;
  uint8_t page, tmp8;
  Si5338Status status;

  if (regIndex > DF_ARRAY_LENGTH(g_registerMasks)) {
    return Si5338StatusInvalidRegister;
  }
  page = (uint8_t)((regIndex >> 8) & 0x1U);
  serializerGetSync(&pSi5338Ctx->pageBitSerializer, &request);
  if (regIndex != SI5338_REGIDX_PAGE && pSi5338Ctx->pageBit != page) {
    status = si5338SetPageBit(pSi5338Ctx, page);
    if (Si5338StatusSuccess != status) {
      /* Force page bit to be changed on next access */
      pSi5338Ctx->pageBit = 0xffU;
      serializerPut(&pSi5338Ctx->pageBitSerializer);
      return status;
    } else {
      pSi5338Ctx->pageBit = page;
    }
  }
  if (mask8 != 0xFFU) {
    status = si5338ReadByteSync(pSi5338Ctx, (uint8_t)regIndex, &tmp8);
    if (Si5338StatusSuccess != status) {
      serializerPut(&pSi5338Ctx->pageBitSerializer);
      return status;
    }
    val8 &= mask8;
    val8 |= tmp8 & ~mask8;
  }
  status = si5338WriteByteSync(pSi5338Ctx, (uint8_t)regIndex, val8);
  serializerPut(&pSi5338Ctx->pageBitSerializer);
  return status;
}

/* I2CCallback invoked when SI5338 register is written */
static void
onRegisterWritten(
  I2cContext* pI2cCtx,
  I2cRequest* pRequest,
  uint8_t data,
  I2cStatus status,
  void* pContextIgnored)
{
  Si5338Request* pSi5338Request = DF_CONTAINER_OF(pRequest, Si5338Request, i2cRequest);
  Si5338DeviceContext* pSi5338Ctx = pSi5338Request->pSi5338Ctx;

  serializerPut(&pSi5338Ctx->pageBitSerializer);
  if (status == I2cStatusSuccess) {
    /* Success */
    pSi5338Request->pCallback(pSi5338Ctx, pSi5338Request, pSi5338Request->data, Si5338StatusSuccess, pSi5338Request->pContext);
  } else {
    /* An I2C error occurred */
    pSi5338Request->pCallback(pSi5338Ctx, pSi5338Request, 0, mapI2cStatus(status), pSi5338Request->pContext);
  }
}

/* I2CCallback invoked when SI5338 register is read */
static void
onRegisterRead(
  I2cContext* pI2cCtx,
  I2cRequest* pRequest,
  uint8_t data,
  I2cStatus status,
  void* pContextIgnored)
{
  Si5338Request* pSi5338Request = DF_CONTAINER_OF(pRequest, Si5338Request, i2cRequest);
  Si5338DeviceContext* pSi5338Ctx = pSi5338Request->pSi5338Ctx;

  serializerPut(&pSi5338Ctx->pageBitSerializer);
  if (status == I2cStatusSuccess) {
    /* Success */
    pSi5338Request->pCallback(pSi5338Ctx, pSi5338Request, data, Si5338StatusSuccess, pSi5338Request->pContext);
  } else {
    /* An I2C error occurred */
    pSi5338Request->pCallback(pSi5338Ctx, pSi5338Request, 0, mapI2cStatus(status), pSi5338Request->pContext);
  }
}

/* I2CCallback invoked when SI5338 register is read as part of read-modify-write cycle */
static void
onRegisterReadForWrite(
  I2cContext* pI2cCtx,
  I2cRequest* pRequest,
  uint8_t data,
  I2cStatus status,
  void* pContextIgnored)
{
  Si5338Request* pSi5338Request = DF_CONTAINER_OF(pRequest, Si5338Request, i2cRequest);
  Si5338DeviceContext* pSi5338Ctx = pSi5338Request->pSi5338Ctx;
  uint16_t regIndex = pSi5338Request->regIndex;
  uint8_t val8;

  if (status == I2cStatusSuccess) {
    /* Perform the write part of the read-modify-write cycle */
    val8 = pSi5338Request->data & pSi5338Request->mask;
    val8 |= data & ~pSi5338Request->mask;
    i2cWrite(pI2cCtx, &pSi5338Request->i2cRequest, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, (uint8_t)regIndex, val8, onRegisterWritten, NULL);
  } else {
    /* An I2C error occurred */
    serializerPut(&pSi5338Ctx->pageBitSerializer);
    pSi5338Request->pCallback(pSi5338Ctx, pSi5338Request, 0, mapI2cStatus(status), pSi5338Request->pContext);
  }
}

/* I2CCallback invoked when SI5338 page register is written */
static void
onPageBitWritten(
  I2cContext* pI2cCtx,
  I2cRequest* pRequest,
  uint8_t data,
  I2cStatus status,
  void* pContextIgnored)
{
  Si5338Request* pSi5338Request = DF_CONTAINER_OF(pRequest, Si5338Request, i2cRequest);
  Si5338DeviceContext* pSi5338Ctx = pSi5338Request->pSi5338Ctx;
  uint16_t regIndex = pSi5338Request->regIndex;
  uint8_t mask;

  if (status == I2cStatusSuccess) {
    if (pSi5338Request->bWrite) {
      mask = pSi5338Request->mask;
      if (mask != 0xFFU) {
        /* Need to first read the register in question, to preserve some bits */
        i2cRead(pI2cCtx, &pSi5338Request->i2cRequest, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, (uint8_t)regIndex, onRegisterReadForWrite, NULL);
      } else {
        /* Can write to the register in question without first reading it */
        i2cWrite(pI2cCtx, &pSi5338Request->i2cRequest, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, (uint8_t)regIndex, pSi5338Request->data, onRegisterWritten, NULL);
      }
    } else {
      /* Read the register in question */
      i2cRead(pI2cCtx, &pSi5338Request->i2cRequest, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, (uint8_t)regIndex, onRegisterRead, NULL);
    }
  } else {
    /* An I2C error occurred */
    serializerPut(&pSi5338Ctx->pageBitSerializer);
    pSi5338Request->pCallback(pSi5338Ctx, pSi5338Request, 0, mapI2cStatus(status), pSi5338Request->pContext);
  }
}

/* SerializerCallback invoked when SI5338 page bit serializer is acquired */
static void
onPageBitAcquired(
  SerializerQueue* pQueue,
  SerializerRequest* pRequest,
  void* pContextIgnored)
{
  Si5338Request* pSi5338Request = DF_CONTAINER_OF(pRequest, Si5338Request, pageBitRequest);
  Si5338DeviceContext* pSi5338Ctx = pSi5338Request->pSi5338Ctx;
  uint16_t regIndex = pSi5338Request->regIndex;
  uint8_t page, mask;

  page = (uint8_t)((regIndex >> 8) & 0x1U);
  if (regIndex != SI5338_REGIDX_PAGE && page != pSi5338Ctx->pageBit) {
    /* Need to change the page bit in the SI5338 */
    i2cWrite(pSi5338Ctx->pI2cCtx, &pSi5338Request->i2cRequest, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, SI5338_REGIDX_PAGE, page, onPageBitWritten, NULL);
  } else {
    if (pSi5338Request->bWrite) {
      mask = pSi5338Request->mask;
      if (mask != 0xFFU) {
        /* Need to first read the register in question, to preserve some bits */
        i2cRead(pSi5338Ctx->pI2cCtx, &pSi5338Request->i2cRequest, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, (uint8_t)regIndex, onRegisterReadForWrite, NULL);
      } else {
        /* Can write to the register in question without first reading it */
        i2cWrite(pSi5338Ctx->pI2cCtx, &pSi5338Request->i2cRequest, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, (uint8_t)regIndex, pSi5338Request->data, onRegisterWritten, NULL);
      }
    } else {
      /* Read the register in question */
      i2cRead(pSi5338Ctx->pI2cCtx, &pSi5338Request->i2cRequest, pSi5338Ctx->busIndex, pSi5338Ctx->deviceIndex, (uint8_t)regIndex, onRegisterRead, NULL);
    }
  }
}

void
si5338Read(
  Si5338DeviceContext* pSi5338Ctx,
  Si5338Request* pRequest,
  uint16_t regIndex,
  Si5338Callback* pCallback,
  void* pContext)
{
  dfDebugPrint(5, ("si5338Read: pSi5338Ctx=%p pRequest=%p regIndex=0x%x pCallback=%p\n", pSi5338Ctx, pRequest, regIndex, pCallback));

  if (regIndex > DF_ARRAY_LENGTH(g_registerMasks)) {
    pCallback(pSi5338Ctx, pRequest, 0, Si5338StatusInvalidRegister, pContext);
  }
  pRequest->pSi5338Ctx = pSi5338Ctx;
  pRequest->pCallback = pCallback;
  pRequest->regIndex = regIndex;
  pRequest->bWrite = FALSE;
  pRequest->pContext = pContext;
  serializerGet(&pSi5338Ctx->pageBitSerializer, &pRequest->pageBitRequest, onPageBitAcquired, NULL);
}

void
si5338WriteMask(
  Si5338DeviceContext* pSi5338Ctx,
  Si5338Request* pRequest,
  uint16_t regIndex,
  uint8_t mask8,
  uint8_t val8,
  Si5338Callback* pCallback,
  void* pContext)
{
  dfDebugPrint(5, ("si5338WriteMask: pSi5338Ctx=%p pRequest=%p regIndex=0x%x mask8=0x%02x val8=0x%02x pCallback=%p\n",
    pSi5338Ctx, pRequest, regIndex, mask8, val8, pCallback));

  if (regIndex > DF_ARRAY_LENGTH(g_registerMasks)) {
    pCallback(pSi5338Ctx, pRequest, 0, Si5338StatusInvalidRegister, pContext);
  }
  pRequest->pSi5338Ctx = pSi5338Ctx;
  pRequest->pCallback = pCallback;
  pRequest->regIndex = regIndex;
  pRequest->mask = mask8;
  pRequest->data = val8;
  pRequest->bWrite = TRUE;
  pRequest->pContext = pContext;
  serializerGet(&pSi5338Ctx->pageBitSerializer, &pRequest->pageBitRequest, onPageBitAcquired, NULL);
}

void
si5338Init(
	I2cContext* pI2cCtx,
  Si5338DeviceContext* pSi5338Ctx,
  uint8_t busIndex,
  uint8_t deviceIndex)
{
  uint8_t val8;
  unsigned int i;

  pSi5338Ctx->pI2cCtx = pI2cCtx;
  pSi5338Ctx->busIndex = busIndex;
  pSi5338Ctx->deviceIndex = deviceIndex;
  serializerQueueInit(&pSi5338Ctx->pageBitSerializer);
  /* Establish the page register */
  si5338ReadByteSync(pSi5338Ctx, SI5338_REGIDX_PAGE, &val8);
  pSi5338Ctx->pageBit = val8 & 0x1U;
  for (i = 0; i < 4; i++) {
    pSi5338Ctx->save[i].bLastValid = FALSE;
  }
}

void
si5338DumpRegisters(
  Si5338DeviceContext* pSi5338Ctx)
{
  uint8_t buf[DF_ARRAY_LENGTH(g_registerMasks)];
  size_t i;

  for (i = 0; i < sizeof(buf); i++) {
    si5338ReadSync(pSi5338Ctx, (uint16_t)i, &buf[i]); /* Cast is safe unless more than 65535 registers in SI5338! */
  }
  dfDebugDumpMemory(0, 0, 16, 1, FALSE, buf, sizeof(buf));
}

boolean_t
si5338Interrogate(
  Si5338DeviceContext* pSi5338Ctx,
  uint32_t fRef,
  uint32_t initFrequencies[4])
{
  unsigned int fOutMin, fOutMax;
  boolean_t bOk = TRUE;
  uint8_t val8, i2cAddr, pfdInSel, pfdInDiv;
  uint32_t p1, p2, p3;
  uint64_t k, fVco, fOut;
  uint32_t fRefDivFactor;
  uint16_t i;

  /* Will correct later */
  pSi5338Ctx->vcoFrequency = 0;
  for (i = 0; i < 4; i++) {
    initFrequencies[i] = 0;
  }

  /* Read I2C address register */
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_I2C_ADDR, &i2cAddr));

  /* Get PFD input select register */
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_PFD_IN_SEL, &val8));
  pfdInSel = (val8 >> 5) & 0x7U;
  pfdInDiv = val8 & 0x3U;

  /* Get MSn_P1, MSn_P2 & MSn_P3 */
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 0, &val8));
  p1 = val8;
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 1, &val8));
  p1 |= (uint32_t)val8 << 8;
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 2, &val8));
  p1 |= (uint32_t)(val8 & 0x3U) << 16;
  p2 = val8 >> 6;
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 3, &val8));
  p2 |= (uint32_t)val8 << 6;
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 4, &val8));
  p2 |= (uint32_t)val8 << 14;
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 5, &val8));
  p2 |= (uint32_t)val8 << 22;
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 6, &val8));
  p3 = (uint32_t)val8 << 0;
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 7, &val8));
  p3 |= (uint32_t)val8 << 8;
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 8, &val8));
  p3 |= (uint32_t)val8 << 16;
  bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSN_P1 + 9, &val8));
  p3 |= (uint32_t)(val8 & 0x3FU) << 24;

  if (!bOk) {
    dfDebugPrint(0, ("*** si5338Interrogate: failed to read one or more SI5338 registers\n"));
    return FALSE;
  }

  if (i2cAddr != pSi5338Ctx->deviceIndex) {
    dfDebugPrint(0, ("*** si5338Interrogate: SI5338 I2C_ADDR=0x%02x, expected 0x%02x\n",
      i2cAddr, pSi5338Ctx->deviceIndex));
    return FALSE;
  }

  switch (pfdInSel) {
  case 0U: /* REFCLK */
  case 4U: /* XOCLK */
    fRefDivFactor = 1U;
    break;

  case 2U: /* DIVREFCLK */
    if (pfdInDiv > 5U) {
      dfDebugPrint(0, ("*** si5338Interrogate: invalid SI5338 PFD reference clock divide select value (%u)\n"));
      return FALSE;
    }
    fRefDivFactor = 1U << pfdInDiv;
    break;

  case 1U: /* FBCLK */
  case 3U: /* DIVFBCLK */
  case 5U: /* NOCLK */
  default:
    dfDebugPrint(0, ("*** si5338Interrogate: invalid SI5338 PFD reference clock select (%u)\n", pfdInSel));
    return FALSE;
  }

  /*
  ** Find fVco from fRef, MSn_P1, MSn_P2 and MSn_P3 by reversing the formulae
  ** for MSn_P1, MSn_P2 and MSn_P3. We have
  **
  **    fVco = fRef * (a + b/c)
  **    p1 = [128(ac + b) div c] - 512
  **    p2 = 128(ac + b) mod c
  **    p3 = c
  **
  ** where p1, p2 and p3 are obtained from SI5338 registers, and fRef is
  ** obtained from VPD.
  **
  ** => 128(ac + b) = p2 + (p1 + 512) * p3
  ** => ac + b = [p2 + (p1 + 512) * p3] / 128
  **
  ** Let k = ac + b = [p2 + (p1 + 512) * p3] / 128, then:
  **
  **    a + b/c = k/c = k/p3
  ** => fVco = fRef * k/p3
  **
  ** So first find k, then find fVco.
  */
  if (0 == p3) {
  	dfDebugPrint(0, ("*** si5338Interrogate: P3 value is zero\n"));
    return FALSE;
  }
  k = ((uint64_t)p3 * (uint64_t)(p1 + 512U) + (uint64_t)p2) >> 7;
  fVco = dfDivide64By64((uint64_t)fRef * k, (uint64_t)p3 * fRefDivFactor);
  if (fVco < g_fVcoMin || fVco > g_fVcoMax) {
  	dfDebugPrint(0, ("+++ si5338Interrogate: calculated Fvco is out of range (%llu Hz)\n", (unsigned long long)fVco));
  } else {
    pSi5338Ctx->vcoFrequency = (uint32_t)fVco;
  	dfDebugPrint(1, ("si5338Interrogate: VCO frequency is %lu Hz\n", (unsigned long)pSi5338Ctx->vcoFrequency));

    /* Find clock output frequencies - cast is safe due to above range check on fVco */
    fOutMin = (uint32_t)fVco / 568; /* Minimum possible output frequency = fVco / 568 */
    fOutMax = (uint32_t)fVco >> 3;  /* Maximum possible output frequency = fVco / 8 */
    for (i = 0; i < 4; i++) {
      /* Get MSn_P1, MSn_P2 & MSn_P3 */
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 0, &val8));
      p1 = val8;
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 1, &val8));
      p1 |= (uint32_t)val8 << 8;
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 2, &val8));
      p1 |= (uint32_t)(val8 & 0x3U) << 16;
      p2 = (val8 >> 2) & 0x3FU;
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 3, &val8));
      p2 |= (uint32_t)val8 << 6;
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 4, &val8));
      p2 |= (uint32_t)val8 << 14;
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 5, &val8));
      p2 |= (uint32_t)val8 << 22;
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 6, &val8));
      p3 = (uint32_t)val8 << 0;
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 7, &val8));
      p3 |= (uint32_t)val8 << 8;
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 8, &val8));
      p3 |= (uint32_t)val8 << 16;
      bOk &= (Si5338StatusSuccess == si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_MSx_P1(i) + 9, &val8));
      p3 |= (uint32_t)(val8 & 0x3FU) << 24;
      dfDebugPrint(2, ("si5338Interrogate: multisynth %u: p1=%u p2=%u p3=%u\n", i, p1, p2, p3));
      k = ((uint64_t)p3 * (uint64_t)(p1 + 512U) + (uint64_t)p2) >> 7;
      fOut = dfDivide64By64(fVco * p3, k);
      if (fOut < fOutMin || fOut > fOutMax) {
  	    dfDebugPrint(0, ("+++ si5338Interrogate: clock output %u frequency is out of range (%llu Hz)\n", i, (unsigned long long)fOut));
      } else {
  	    dfDebugPrint(1, ("si5338Interrogate: clock output %u frequency is %lu Hz\n", i, (unsigned long)fOut));
        initFrequencies[i] = (uint32_t)fOut; /* Cast is safe due to above range check on fOut */
      }
    }
  }

  return TRUE;
}

boolean_t
si5338ConfigureFull(
  Adb3CoreDeviceContext* pDevCtx,
  Si5338DeviceContext* pSi5338Ctx,
  uint32_t start,  /* VPD space start address */
  uint32_t length) /* VPD space length (bytes) */
{
  const unsigned int pllPollIntervalUs = 25000; /* 25 ms poll interval waiting for PLL lock */
  const unsigned int maxCount = 10;             /* Timeout is 10 periods of poll interval waiting for PLL lock */
  Si5338ConfigurationRecord record;
  unsigned int numRecord = 0;
  CoreVpdStatus vpdStatus;
  Si5338Status si5338Status;
  uint16_t addr, tmp16;
  uint8_t mask8, val8, fcal0, fcal1, fcal2;
  unsigned int count;

  vpdStatus = pDevCtx->methods.pReadWriteVpd(pDevCtx, FALSE, start, sizeof(tmp16), &tmp16);
  if (CoreVpdSuccess != vpdStatus) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to probe for valid SI5338 configuration in VPD\n"));
    return FALSE;
  }
  if (tmp16 == 0xFFFFU) {
    dfDebugPrint(1, ("+++ si5338ConfigureFull: valid SI5338 configuration not found in VPD\n"));
    return FALSE;
  }

  /* Disable all outputs */
  val8 = mask8 = 0x10U;
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_DRIVER_DISABLE, mask8, val8);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, val8, SI5338_REGIDX_DRIVER_DISABLE));
    return FALSE;
  }

  /* Pause loss-of-lock detection */
  val8 = 0xE5U;
  mask8 = 0xFFU;
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_LOL_PAUSE, mask8, val8);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, val8, SI5338_REGIDX_LOL_PAUSE));
    return FALSE;
  }

  /* Process SI5338 configuration records that are stored in VPD */
  while (length >= sizeof(record)) {
    /* Read the record header */
    vpdStatus = pDevCtx->methods.pReadWriteVpd(pDevCtx, FALSE, start, sizeof(record), &record);
    if (CoreVpdSuccess != vpdStatus) {
      dfDebugPrint(0, ("*** si5338ConfigureFull: failed to read record of %lu byte(s) from VPD address 0x%lx\n",
        (unsigned long)sizeof(record), (unsigned long)start));
      return FALSE;
    }
    start += sizeof(record);
    addr = (uint16_t)record.addrLo | ((uint16_t)record.addrHi << 8);
    if (addr == 0xFFFFU) {
      break;
    }
    si5338Status = si5338WriteMaskSync(pSi5338Ctx, addr, record.mask & g_registerMasks[addr], record.data);
    if (Si5338StatusSuccess != si5338Status) {
      dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
        record.data, record.mask, addr));
      return FALSE;
    }
    numRecord++;
  }
  dfDebugPrint(1, ("si5338ConfigureFull: processed %u record(s)\n", numRecord));

  /* Validate input clock status */
  for (count = 0;; count++) {
    /* Check for input clock valid */
    si5338Status = si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_LOL_STATUS, &val8);
    if (Si5338StatusSuccess != si5338Status) {
      dfDebugPrint(0, ("*** si5338ConfigureFull: failed to read SI5338 register 0x%x\n", SI5338_REGIDX_LOL_STATUS));
      return FALSE;
    }
    if ((val8 & 0x04U) == 0U) {
      /* Input clock is valid */
      break;
    }
    if (count > maxCount) {
      break;
    }
    dfDelayThreadFor(dfMicrosecondsToTime(pllPollIntervalUs));
  }
  if (count > maxCount) {
    /* Timeout waiting for input clock valid */
    dfDebugPrint(0, ("*** si5338ConfigureFull: timeout waiting for input clock valid\n"));
    return FALSE;
  }

  /* Configure PLL for locking */
  val8 = 0x00U;
  mask8 = 0x80U;
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_FCAL_OVRD_EN, mask8, val8);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, val8, SI5338_REGIDX_FCAL_OVRD_EN));
    return FALSE;
  }

  /* Initiate locking of PLL */
  val8 = 0x02U;
  mask8 = 0xFFU; /* SI5338 AN411 says don't use read-modify-write for this register */
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_PLL_SOFT_RESET, mask8, val8);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, val8, SI5338_REGIDX_PLL_SOFT_RESET));
    return FALSE;
  }

  /* Resume loss-of-lock detection */
  val8 = 0x65U;
  mask8 = 0xFFU;
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_LOL_PAUSE, mask8, val8);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, val8, SI5338_REGIDX_LOL_PAUSE));
    return FALSE;
  }

  for (count = 0; count < maxCount; count++) {
    /* SI5338 datasheet says wait 25 ms */
    dfDelayThreadFor(dfMicrosecondsToTime(pllPollIntervalUs));

    /* Check for PLL lock */
    si5338Status = si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_LOL_STATUS, &val8);
    if (Si5338StatusSuccess != si5338Status) {
      dfDebugPrint(0, ("*** si5338ConfigureFull: failed to read SI5338 register 0x%x\n", SI5338_REGIDX_LOL_STATUS));
      return FALSE;
    }
    if ((val8 & 0x15U) == 0U) {
      /* PLL locked, input clock valid, device callibration not in progress */
      break;
    }
    dfDebugPrint(2, ("si5338ConfigureFull: LOL_STATUS=0x%02x\n", val8));
  }
  if (count == maxCount) {
    /* Timeout waiting for PLL lock */
    dfDebugPrint(0, ("*** si5338ConfigureFull: timeout waiting for PLL lock, LOL_STATUS=0x%02x\n", val8));
    return FALSE;
  }

  /* Copy FCAL registers to FCAL_OVRD registers */
  si5338Status = si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_FCAL2, &fcal2);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to read SI5338 register 0x%x\n", SI5338_REGIDX_FCAL2));
    return FALSE;
  }
  si5338Status = si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_FCAL1, &fcal1);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to read SI5338 register 0x%x\n", SI5338_REGIDX_FCAL1));
    return FALSE;
  }
  si5338Status = si5338ReadSync(pSi5338Ctx, SI5338_REGIDX_FCAL0, &fcal0);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to read SI5338 register 0x%x\n", SI5338_REGIDX_FCAL0));
    return FALSE;
  }
  /* Datasheet says set bits 7:2 to 000101 */
  fcal2 &= 0x03U;
  fcal2 |= 0x14U;
  mask8 = 0xFFU;
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_FCAL_OVRD2, mask8, fcal2);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, fcal2, SI5338_REGIDX_FCAL_OVRD2));
    return FALSE;
  }
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_FCAL_OVRD1, mask8, fcal1);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, fcal1, SI5338_REGIDX_FCAL_OVRD1));
    return FALSE;
  }
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_FCAL_OVRD0, mask8, fcal0);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, fcal0, SI5338_REGIDX_FCAL_OVRD0));
    return FALSE;
  }

  /* Set FCAL_OVRD_EN enable bit */
  val8 = mask8 = 0x80U;
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_FCAL_OVRD_EN, mask8, val8);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, val8, SI5338_REGIDX_FCAL_OVRD_EN));
    return FALSE;
  }

  /* Reenable all outputs */
  mask8 = 0x10U;
  val8 = 0x00U;
  si5338Status = si5338WriteMaskSync(pSi5338Ctx, SI5338_REGIDX_DRIVER_DISABLE, mask8, val8);
  if (Si5338StatusSuccess != si5338Status) {
    dfDebugPrint(0, ("*** si5338ConfigureFull: failed to write value 0x%02x (mask 0x%02x) to SI5338 register 0x%x\n",
      mask8, val8, SI5338_REGIDX_DRIVER_DISABLE));
    return FALSE;
  }

  dfDebugPrint(1, ("si5338ConfigureFull: successful\n"));

  return TRUE;
}

static Si5338ProgramStatus
mapSi5338Status(
  Si5338Status status)
{
  switch (status) {
  case Si5338StatusSuccess:
    return Si5338ProgramStatusSuccess;
  
  case Si5338StatusI2CTimeout:
  case Si5338StatusI2CError:
    return Si5338ProgramStatusI2CError;

  case Si5338StatusInvalidRegister:
  case Si5338StatusGeneralFailure:
  default:
    return Si5338ProgramStatusGeneralFailure;
  }
}

/* SI5338Callback invoked when an SI5338 register access is completed */
static void
onSI5338Written(
  Si5338DeviceContext* pSi5338Ctx,
  Si5338Request* pRequest,
  uint8_t data,
  Si5338Status status,
  void* pContextIgnored)
{
  Si5338ProgramRequest* pProgRequest = DF_CONTAINER_OF(pRequest, Si5338ProgramRequest, request);
  static const uint16_t msAddrs[4] = {
    SI5338_REGIDX_MS0_P1,
    SI5338_REGIDX_MS1_P1,
    SI5338_REGIDX_MS2_P1,
    SI5338_REGIDX_MS3_P1
  };
  unsigned int stage = pProgRequest->stage;
  unsigned int msIndex = pProgRequest->msIndex;
  CoreClockWord* pClockWord = &pProgRequest->clockWord;
  uint16_t msAddr;
  uint8_t enableBit, val8;

  dfDebugPrint(4, ("onSI5338Written: pRequest=%p status=%u\n", pRequest, status));

  if (Si5338StatusSuccess != status) {
    /* An error occurred while writing to a SI5338 register */
    pProgRequest->pCallback(pSi5338Ctx, pProgRequest, pProgRequest->msIndex, mapSi5338Status(status), pProgRequest->pContext);
    return;
  }

  /* Look up address of selected MultiSynth in SI5338 */
  stage = pProgRequest->stage;
  msIndex = pProgRequest->msIndex;
  pClockWord = &pProgRequest->clockWord;
  msAddr = msAddrs[msIndex];
  pProgRequest->stage = stage + 1;

  dfDebugPrint(4, ("onSI5338Written: stage=%u msIndex=%u msAddr=0x%x\n", stage, msIndex, msAddr));

  switch (stage) {
  case 0:
    /* Set MSx_P1[7:0] */
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 0, 0xFFU, (uint8_t)(pClockWord->d0 >> 0), onSI5338Written, NULL);
    break;

  case 1:
    /* Set MSx_P1[15:8] */
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 1, 0xFFU, (uint8_t)(pClockWord->d0 >> 8), onSI5338Written, NULL);
    break;

  case 2:
    /* Set MSx_P1[17:16], MSx_P2[5:0] */
    val8 = (uint8_t)((pClockWord->d0 >> 16) & 0x3U);
    val8 |= (uint8_t)((pClockWord->d1 & 0x3FU) << 2);
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 2, 0xFFU, val8, onSI5338Written, NULL);
    break;

  case 3:
    /* Set MSx_P2[13:16] */
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 3, 0xFFU, (uint8_t)(pClockWord->d1 >> 6), onSI5338Written, NULL);
    break;

  case 4:
    /* Set MSx_P2[21:14] */
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 4, 0xFFU, (uint8_t)(pClockWord->d1 >> 14), onSI5338Written, NULL);
    break;

  case 5:
    /* Set MSx_P2[29:22] */
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 5, 0xFFU, (uint8_t)((pClockWord->d1 >> 22) & 0x3FU), onSI5338Written, NULL);
    break;

  case 6:
    /* Set MSx_P3[7:0] */
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 6, 0xFFU, (uint8_t)(pClockWord->d2 >> 0), onSI5338Written, NULL);
    break;

  case 7:
    /* Set MSx_P3[15:8] */
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 7, 0xFFU, (uint8_t)(pClockWord->d2 >> 8), onSI5338Written, NULL);
    break;

  case 8:
    /* Set MSx_P3[23:16] */
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 8, 0xFFU, (uint8_t)(pClockWord->d2 >> 16), onSI5338Written, NULL);
    break;

  case 9:
    /* Set MSx_P3[29:24] */
    si5338WriteMask(pSi5338Ctx, pRequest, msAddr + 9, 0x3FU, (uint8_t)(pClockWord->d2 >> 24), onSI5338Written, NULL);
    break;

  case 10:
    /* Reenable the selected clock output */
    enableBit = (uint8_t)(0x1U << msIndex);
    si5338WriteMask(pSi5338Ctx, pRequest, SI5338_REGIDX_DRIVER_DISABLE, enableBit, 0, onSI5338Written, NULL);
    break;

  case 11:
    /* Successfully wrote all necessary SI5338 registers */
    pSi5338Ctx->save[pProgRequest->msIndex].bLastValid = TRUE;
    pSi5338Ctx->save[pProgRequest->msIndex].last = pProgRequest->clockWord;
    pProgRequest->pCallback(pSi5338Ctx, pProgRequest, pProgRequest->msIndex, Si5338ProgramStatusSuccess, pProgRequest->pContext);
    break;

  default:
    /* Should never get here! */
    pProgRequest->pCallback(pSi5338Ctx, pProgRequest, pProgRequest->msIndex, Si5338ProgramStatusGeneralFailure, pProgRequest->pContext);
    break;
  }
}

Si5338ProgramStatus
si5338ClockProgram(
  Si5338DeviceContext* pSi5338Ctx,
  Si5338ProgramRequest* pRequest,
  unsigned int msIndex,
  const CoreClockWord* pClockWord,
  Si5338ProgramCallback* pCallback,
  void* pContext)
{
  uint8_t enableBit;

  dfDebugPrint(4, ("si5338ClockProgram: pRequest=%p msIndex=%u pCallback=%p\n", pRequest, msIndex, pCallback));

  if (msIndex > 3) {
    return Si5338ProgramStatusInvalidIndex;
  }

  /* Start programming the SI5338 */
  pRequest->msIndex = msIndex;
  pRequest->stage = 0;
  pRequest->clockWord = *pClockWord;
  pRequest->pCallback = pCallback;
  pRequest->pContext = pContext;

  /* First, disable the selected clock output */
  enableBit = (uint8_t)(0x1U << msIndex);
  si5338WriteMask(pSi5338Ctx, &pRequest->request, SI5338_REGIDX_DRIVER_DISABLE, enableBit, enableBit, onSI5338Written, NULL);

  return Si5338ProgramStatusSuccess;
}

static void
onProgramDone(
  Si5338DeviceContext* pSi5338Ctx,
  Si5338ProgramRequest* pRequest,
  unsigned int msIndex,
  Si5338ProgramStatus status,
  void* pContextIgnored)
{
  Si5338ProgramRequestSync* pRequestSync = DF_CONTAINER_OF(pRequest, Si5338ProgramRequestSync, request);

  dfDebugPrint(4, ("onProgramDone: msIndex=%u status=%u\n", msIndex, status));

  pRequestSync->status = status;
  dfEventSignal(&pRequestSync->ev);
}

Si5338ProgramStatus
si5338ClockProgramSync(
  Si5338DeviceContext* pSi5338Ctx,
  unsigned int msIndex, /* Multisynth index */
  const CoreClockWord* pClockWord)
{
  Si5338ProgramRequestSync req;
  Si5338ProgramStatus status;

  dfDebugPrint(3, ("si5338ClockProgramSync: msIndex=%u\n", msIndex));

  dfEventInit(&req.ev);
  status = si5338ClockProgram(pSi5338Ctx, &req.request, msIndex, pClockWord, onProgramDone, NULL);
  if (Si5338ProgramStatusSuccess == status) {
    dfEventWait(&req.ev);
    status = req.status;
  }
  dfEventUninit(&req.ev);

  return status;
}

void
si5338SaveState(
  Si5338DeviceContext* pSi5338Ctx)
{
  /* Invalidate the SI5338 page bit shadow to force the page register to be reinitialized on restore */
  pSi5338Ctx->pageBit = 0xFFU;
}

Si5338ProgramStatus
si5338RestoreState(
  Si5338DeviceContext* pSi5338Ctx)
{
  Si5338ProgramStatus status = Si5338ProgramStatusSuccess;
  unsigned int msIndex;

  for (msIndex = 0; msIndex < 4; msIndex++) {
    if (pSi5338Ctx->save[msIndex].bLastValid) {
      dfDebugPrint(2, ("si5338RestoreState: restoring SI5338 multisynth %u\n", msIndex));
      status = si5338ClockProgramSync(pSi5338Ctx, msIndex, &pSi5338Ctx->save[msIndex].last);
      if (Si5338ProgramStatusSuccess != status) {
        return status;
      }
    }
  }

  return status;
}

ClockProgramStatus
si5338MapProgramStatus(
  Si5338ProgramStatus status)
{
  switch (status) {
  case Si5338ProgramStatusSuccess:
    return ClockProgramSuccess;

  case Si5338ProgramStatusI2CError:
    return ClockProgramHardwareError;

  case Si5338ProgramStatusInvalidIndex:
    return ClockProgramInvalidIndex;

  case Si5338ProgramStatusGeneralFailure:
  default:
    return ClockProgramGeneralFailure;
  }
}

ClockWordStatus
si5338MapClockWordStatus(
  Si5338ClockWordStatus status)
{
  switch (status) {
  case Si5338ClockWordStatusSuccess:
    return ClockWordSuccess;

  case Si5338ClockWordStatusOutOfRange:
    return ClockWordOutOfRange;

  case Si5338ClockWordStatusGeneralFailure:
  default:
    return ClockWordGeneralFailure;
  }
}
